/*
	STM32F0DISCOVERY tutorial
	TIM3 PWM LED control by ADC temperature sensor
	gbm, 12'2012
*/

#include "stm32f0xx.h"
//========================================================================
// defs for STM32F05x chips
#define GPIO_MODER_OUT	1
#define GPIO_MODER_AF	2

#define TIM_CCMR2_OC3M_PWM1	0x0060	// OC3M[2:0] - PWM mode 1
#define TIM_CCMR2_OC4M_PWM1	0x6000	// OC4M[2:0] - PWM mode 1

#define ADC_SMPR_71_5	6

typedef __IO uint32_t * __IO32p;
//========================================================================
// defs for STM32F0DISCOVERY board
#define LED_PORT		GPIOC
#define BLUE_LED_BIT	8
#define GREEN_LED_BIT	9

#define BUTTON_PORT		GPIOA
#define BUTTON_BIT		0

#define BLUE_LED_PWM	TIM3->CCR3
#define GREEN_LED_PWM	TIM3->CCR4
//========================================================================
// Timings
#define SYSCLK_FREQ	HSI_VALUE

// PWM constants
#define PWM_FREQ	400	// Hz
#define PWM_STEPS	50
#define	PWM_CLK		SYSCLK_FREQ
#define	PWM_PRE		(PWM_CLK / PWM_FREQ / PWM_STEPS)
#define BLUE_LED_PWM	TIM3->CCR3
#define GREEN_LED_PWM	TIM3->CCR4
#define LED_MAX	(PWM_STEPS - 1)
#define LED_DIM	1

// temperature sensor parms
#define LONG_AVG_SHIFT	4
#define SHORT_AVG_SHIFT	8
#define T_DELTA	2
//========================================================================
struct init_entry_ {
	volatile uint32_t *loc;
	uint32_t value;
};

static __INLINE void writeregs(const struct init_entry_ *p)
{
	for (; p->loc; p ++)
		*p->loc = p->value;
}
//========================================================================
void SystemInit(void)
{
	FLASH->ACR = FLASH_ACR_PRFTBE;	// enable prefetch
}
//========================================================================
static const struct init_entry_ init_table[] = 
{
	// port setup
	{&RCC->AHBENR, RCC_AHBENR_GPIOCEN |  RCC_AHBENR_GPIOAEN},	// GPIOC, GPIOA
	{&LED_PORT->MODER, GPIO_MODER_AF << (GREEN_LED_BIT << 1)
		| GPIO_MODER_AF << (BLUE_LED_BIT << 1)},	// set LED pins as AF
	{&RCC->AHBENR, RCC_AHBENR_GPIOAEN},	// GPIOA
	// ADC setup
	{&RCC->APB2ENR, RCC_APB2ENR_ADC1EN},	// ADC
	{&ADC1->CHSELR, ADC_CHSELR_CHSEL16},	// ADC
	{&ADC1->SMPR, ADC_SMPR_71_5},	// ADC
	{&ADC->CCR, ADC_CCR_TSEN | ADC_CCR_VREFEN},	// ADC
	{&ADC1->CFGR1, ADC_CFGR1_WAIT | ADC_CFGR1_CONT},	// ADC
	// ADC cal
	{&ADC1->CR, ADC_CR_ADCAL},	// ADC
	// PWM timer setup - TIM3
	{&RCC->APB1ENR, 1 << 1},	// TIM3
	{(__IO32p)&TIM3->PSC, PWM_PRE - 1},
	{(__IO32p)&TIM3->ARR, PWM_STEPS - 1},
	// blue - CH3, green - CH4
	{(__IO32p)&TIM3->CCMR2, TIM_CCMR2_OC4M_PWM1 | TIM_CCMR2_OC4PE
		| TIM_CCMR2_OC3M_PWM1 | TIM_CCMR2_OC3PE},	// PWM mode 1, buffered preload
	{(__IO32p)&TIM3->CCER, TIM_CCER_CC4E | TIM_CCER_CC3E},	// enable CH3, 4 output
	{(__IO32p)&TIM3->DIER, TIM_DIER_UIE},	// enable update interrupt
	{(__IO32p)&TIM3->CR1, TIM_CR1_ARPE | TIM_CR1_CEN},	// auto reload buffer, enable
	// interrupts and sleep
	{&NVIC->ISER[0], 1 << TIM3_IRQn},	// enable interrupt
	{&SCB->SCR, SCB_SCR_SLEEPONEXIT_Msk},	// sleep while not in handler
	{0, 0}
};

//========================================================================
int main(void)
{
	writeregs(init_table);
	__WFI();	// go to sleep
}

//========================================================================
void TIM3_IRQHandler(void)
{
	static uint8_t tdiv = 0;
	static uint8_t blue_target = LED_DIM, green_target = LED_DIM;
	static uint32_t long_avg, short_avg;
	uint32_t val;

	TIM3->SR = ~TIM_SR_UIF;	// clear interrupt flag

	if ((++ tdiv & 3) == 0)
	{
		if (ADC1->ISR & ADC_ISR_EOC)
		{
			val = ADC1->DR;	// 0 before first conversion
		
			if (short_avg == 0)
			{
				// initial measure - set
				short_avg = val << SHORT_AVG_SHIFT;
				long_avg = val << LONG_AVG_SHIFT;
			}
			else
			{
				// low-pass filters
				short_avg = short_avg + val - (short_avg >> SHORT_AVG_SHIFT);
				long_avg = long_avg + val - (long_avg >> LONG_AVG_SHIFT);
				if ((short_avg >> SHORT_AVG_SHIFT) == (long_avg >> LONG_AVG_SHIFT))
				{
					// stable
					blue_target = LED_DIM;
					green_target = LED_DIM;
				}
				else if ((short_avg >> SHORT_AVG_SHIFT) > (long_avg >> LONG_AVG_SHIFT) + T_DELTA)
				{
					// warming up
					blue_target = LED_DIM;
					green_target = LED_MAX;
				}
				else if ((short_avg >> SHORT_AVG_SHIFT) + T_DELTA < (long_avg >> LONG_AVG_SHIFT))
				{
					// cooling
					blue_target = LED_MAX;
					green_target = LED_DIM;
				}
			}
		}
		else if (ADC1->ISR & ADC_ISR_ADRDY)
		{
			// ready for conversion
			ADC1->CR = ADC_CR_ADSTART | ADC_CR_ADEN;	// start cont. conversion
		}
		else if ((ADC1->CR & (ADC_CR_ADCAL | ADC_CR_ADEN)) == 0)
		{
			// calibrated but not enabled yet - enable
			ADC1->CR = ADC_CR_ADEN;
		}
	}
	
	if ((val = BLUE_LED_PWM) != blue_target)
		BLUE_LED_PWM = val < blue_target ? val + 1 : val - 1;
			
	if ((val = GREEN_LED_PWM) != green_target)
		GREEN_LED_PWM = val < green_target ? val + 1 : val - 1;
}
