/*
	STM32F0DISCOVERY tutorial
	SysTick-based blinker with table-driven init and blink mode switching
	gbm, 12'2012
*/

#include "stm32f0xx.h"
//========================================================================
// defs for STM32F0DISCOVERY board

#define LED_PORT		GPIOC
#define BLUE_LED_BIT	8
#define GREEN_LED_BIT	9

#define BUTTON_PORT		GPIOA
#define BUTTON_BIT		0

#define GPIO_MODER_OUT	1

#define SYSCLK_FREQ	HSI_VALUE

#define SYSTICK_FREQ	100	// 100 Hz -> 10 ms
#define BLINK_PERIOD	50	// * 10 ms
#define BLUE_ACK_TIME	100
//========================================================================
struct init_entry_ {
	volatile uint32_t *loc;
	uint32_t value;
};

static __INLINE void writeregs(const struct init_entry_ *p)
{
	for (; p->loc; p ++)
		*p->loc = p->value;
}

//========================================================================
void SystemInit(void)
{
	FLASH->ACR = FLASH_ACR_PRFTBE;	// enable prefetch
}
//========================================================================
static const struct init_entry_ init_table[] = 
{
	// port setup
	{&RCC->AHBENR, RCC_AHBENR_GPIOCEN | RCC_AHBENR_GPIOAEN},	// GPIOC, GPIOA
	{&LED_PORT->MODER, GPIO_MODER_OUT << (GREEN_LED_BIT << 1)
		| GPIO_MODER_OUT << (BLUE_LED_BIT << 1)},	// set LED pins as outputs
	//SysTick setup
	{&SysTick->LOAD, SYSCLK_FREQ / SYSTICK_FREQ - 1},
	{&SysTick->VAL, 0},
	{&SysTick->CTRL, SysTick_CTRL_CLKSOURCE_Msk | SysTick_CTRL_TICKINT_Msk
		| SysTick_CTRL_ENABLE_Msk},
	{&SCB->SCR, SCB_SCR_SLEEPONEXIT_Msk},	// sleep while not in handler
	{0, 0}
};

//========================================================================
int main(void)
{
	writeregs(init_table);
	__WFI();	// go to sleep
}

//========================================================================
void SysTick_Handler(void)
{
	static enum {BM_SLOW, BM_FAST, BM_OFF} blink_mode = BM_SLOW;
	static uint8_t blink_timer = BLINK_PERIOD;
	static uint8_t blue_led_timer = 0, green_on_time = 0;
	static uint8_t bstate = 0;
	static const uint8_t blink_period[] = {100, 50, 1};

	bstate = (bstate << 1 & 0xf) | (BUTTON_PORT->IDR >> BUTTON_BIT & 1);

	if (blue_led_timer)
	{
		if (-- blue_led_timer == 0)
			LED_PORT->BRR = 1 << BLUE_LED_BIT;	// blue OFF
	}
	else if (bstate == 1)
	{
		// button was released, now it is pressed - change blink mode
		if (++ blink_mode > BM_OFF)
			blink_mode = BM_SLOW;
		blue_led_timer = BLUE_ACK_TIME;
		LED_PORT->BSRR = 1 << BLUE_LED_BIT;	// blue ON
	}
	
	if (blue_led_timer && -- blue_led_timer == 0)
		LED_PORT->BRR = 1 << BLUE_LED_BIT;	// blue OFF
	
	if (-- blink_timer == 0)
	{
		blink_timer = blink_period[blink_mode];
		green_on_time = blink_timer >> 1;
		LED_PORT->BRR = 1 << GREEN_LED_BIT;	// green OFF
	}
	else if (blink_timer == green_on_time)
		LED_PORT->BSRR = 1 << GREEN_LED_BIT;	// green ON
}
