/*
	STM32F0DISCOVERY tutorial
	TIM3 PWM blinker with table-driven init and blink mode switching
	gbm, 12'2012
*/

#include "stm32f0xx.h"
//========================================================================
// defs for STM32F05x chips
#define GPIO_MODER_OUT	1
#define GPIO_MODER_AF	2

#define TIM_CCMR2_OC3M_PWM1	0x0060	// OC3M[2:0] - PWM mode 1
#define TIM_CCMR2_OC4M_PWM1	0x6000	// OC4M[2:0] - PWM mode 1

typedef __IO uint32_t * __IO32p;
//========================================================================
// defs for STM32F0DISCOVERY board
#define LED_PORT		GPIOC
#define BLUE_LED_BIT	8
#define GREEN_LED_BIT	9

#define BUTTON_PORT		GPIOA
#define BUTTON_BIT		0

#define BLUE_LED_PWM	TIM3->CCR3
#define GREEN_LED_PWM	TIM3->CCR4
//========================================================================
// Timings
#define SYSCLK_FREQ	HSI_VALUE
#define BLINK_PERIOD	50	// * 10 ms
#define BTN_ACK_PERIOD	50	// * 10 ms

// PWM constants
#define PWM_FREQ	400	// Hz
#define PWM_STEPS	80
#define	PWM_CLK		SYSCLK_FREQ
#define	PWM_PRE		(PWM_CLK / PWM_FREQ / PWM_STEPS)
#define LED_MAX	(PWM_STEPS - 1)
#define LED_DIM	1
#define LED_OFF	0
//========================================================================
struct init_entry_ {
	volatile uint32_t *loc;
	uint32_t value;
};

static __INLINE void writeregs(const struct init_entry_ *p)
{
	for (; p->loc; p ++)
		*p->loc = p->value;
}
//========================================================================
void SystemInit(void)
{
	FLASH->ACR = FLASH_ACR_PRFTBE;	// enable prefetch
}
//========================================================================
static const struct init_entry_ init_table[] = 
{
	// port setup
	{&RCC->AHBENR, RCC_AHBENR_GPIOCEN | RCC_AHBENR_GPIOAEN},	// GPIOC, GPIOA
	{&LED_PORT->MODER, GPIO_MODER_AF << (GREEN_LED_BIT << 1)
		| GPIO_MODER_AF << (BLUE_LED_BIT << 1)},	// set LED pins as AF
	{&RCC->AHBENR, RCC_AHBENR_GPIOAEN},	// GPIOA
	// PWM timer setup - TIM3
	{&RCC->APB1ENR, RCC_APB1ENR_TIM3EN},	// TIM3
	{(__IO32p)&TIM3->PSC, PWM_PRE - 1},
	{(__IO32p)&TIM3->ARR, PWM_STEPS - 1},
	// blue - CH3, green - CH4
	{(__IO32p)&TIM3->CCMR2, TIM_CCMR2_OC4M_PWM1 | TIM_CCMR2_OC4PE
		| TIM_CCMR2_OC3M_PWM1 | TIM_CCMR2_OC3PE},	// PWM mode 1, buffered preload
	{(__IO32p)&TIM3->CCER, TIM_CCER_CC4E | TIM_CCER_CC3E},	// enable CH3, 4 output
	{(__IO32p)&TIM3->DIER, TIM_DIER_UIE},	// enable update interrupt
	{(__IO32p)&TIM3->CR1, TIM_CR1_ARPE | TIM_CR1_CEN},	// auto reload, enable
	// interrupts and sleep
	{&NVIC->ISER[0], 1 << TIM3_IRQn},	// enable interrupt
	{&SCB->SCR, SCB_SCR_SLEEPONEXIT_Msk},	// sleep while not in handler
	{0, 0}
};

//========================================================================
int main(void)
{
	writeregs(init_table);
	__WFI();	// go to sleep
}

//========================================================================
void TIM3_IRQHandler(void)
{
	static uint8_t blink_timer = BLINK_PERIOD;
	static enum {BM_SLOW, BM_FAST, BM_OFF} blink_mode = BM_SLOW;
	static uint8_t bstate = 0;
	static uint8_t blue_led_timer = 0, on_time = 0;
	static const uint8_t blink_periods[] = {100, 50, 1};
	static uint8_t tdiv = 0;
	static uint8_t blue_target = LED_DIM, green_target = LED_DIM;

	uint32_t pwmval;

	TIM3->SR = ~TIM_SR_UIF;	// clear interrupt flag

	if ((++ tdiv & 3) == 0)
	{
		bstate = (bstate << 1 & 0xf) | (BUTTON_PORT->IDR >> BUTTON_BIT & 1);
		
		if (blue_led_timer)
		{
			if (-- blue_led_timer == 0)
				blue_target = LED_DIM;
		}
		else if (bstate == 1)
		{
			// button was released, now it is pressed - change blink mode
			if (++ blink_mode > BM_OFF)
				blink_mode = BM_SLOW;
			blue_led_timer = BTN_ACK_PERIOD;
			blue_target = LED_MAX;
		}
		
		if (-- blink_timer == 0)
		{
			blink_timer = blink_periods[blink_mode];
			on_time = blink_timer >> 1;
			green_target = LED_DIM;
		}
		else if (blink_timer == on_time)
			green_target = LED_MAX;
	}
	
	if ((pwmval = BLUE_LED_PWM) != blue_target)
		BLUE_LED_PWM = pwmval < blue_target ? pwmval + 1 : pwmval - 1;
			
	if ((pwmval = GREEN_LED_PWM) != green_target)
		GREEN_LED_PWM = pwmval < green_target ? pwmval + 1 : pwmval - 1;
}
