/*
	STM32F0DISCOVERY tutorial
	TIM6 software-PWM blinker with table-driven init and blink mode switching
	gbm, 01'2013
*/

#include "stm32f0xx.h"

//#define SPOIL	// uncomment to show the effect of non-interlocked port update
//#define REPAIR	// uncomment to correct the problem
//========================================================================
// defs for STM32F05x chips
#define GPIO_MODER_OUT	1
#define GPIO_MODER_AF	2

#define TIM_CCMR2_OC3M_PWM1	0x0060	// OC3M[2:0] - PWM mode 1
#define TIM_CCMR2_OC4M_PWM1	0x6000	// OC4M[2:0] - PWM mode 1
typedef __IO uint32_t * __IO32p;
//========================================================================
// defs for STM32F0DISCOVERY board
#define LED_PORT		GPIOC
#define BLUE_LED_BIT	8
#define GREEN_LED_BIT	9

#define BUTTON_PORT		GPIOA
#define BUTTON_BIT		0

#define BLUE_LED_PWM	TIM3->CCR3
#define GREEN_LED_PWM	TIM3->CCR4
//========================================================================
// Timings
#define SYSCLK_FREQ	HSI_VALUE

#define BLINK_PERIOD	50	// * 10 ms
#define BTN_ACK_PERIOD	50	// * 10 ms

// PWM constants
#define PWM_FREQ	400	// Hz
#define PWM_STEPS	40

#define SOFTPWM_INT_FREQ	(PWM_FREQ * PWM_STEPS)

#define	PWMTMR_CLK		SYSCLK_FREQ
#define	PWMTMR_PRE		(PWMTMR_CLK / SOFTPWM_INT_FREQ / 65536 + 1)
#define	PWMTMR_PERIOD	(PWMTMR_CLK / PWMTMR_PRE / SOFTPWM_INT_FREQ)

#define LED_MAX	(PWM_STEPS - 1)
#define LED_DIM	1
#define LED_OFF	0
//========================================================================
struct init_entry_ {
	volatile uint32_t *loc;
	uint32_t value;
};

static __INLINE void writeregs(const struct init_entry_ *p)
{
	for (; p->loc; p ++)
		*p->loc = p->value;
}
//========================================================================
void SystemInit(void)
{
	FLASH->ACR = FLASH_ACR_PRFTBE;	// enable prefetch
}
//========================================================================
static const struct init_entry_ init_table[] = 
{
	// port setup
	{&RCC->AHBENR, RCC_AHBENR_GPIOCEN | RCC_AHBENR_GPIOAEN},	// GPIOC, GPIOA
	{&LED_PORT->MODER, GPIO_MODER_OUT << (GREEN_LED_BIT << 1)
		| GPIO_MODER_OUT << (BLUE_LED_BIT << 1)},	// set LED pins as AF
	// soft PWM timer setup - TIM6
	{&RCC->APB1ENR, RCC_APB1ENR_TIM6EN},	// TIM6
	{(__IO32p)&TIM6->PSC, PWMTMR_PRE - 1},
	{(__IO32p)&TIM6->ARR, PWMTMR_PERIOD - 1},
	{(__IO32p)&TIM6->DIER, TIM_DIER_UIE},	// enable update interrupt
	{(__IO32p)&TIM6->CR1, TIM_CR1_ARPE | TIM_CR1_CEN},	// auto reload, enable
	// interrupts and sleep
	{&NVIC->ISER[0], 1 << TIM6_DAC_IRQn},	// enable interrupt
#ifndef SPOIL
	{&SCB->SCR, SCB_SCR_SLEEPONEXIT_Msk},	// sleep while not in handler
#endif
	{0, 0}
};
//========================================================================
int main(void)
{
	writeregs(init_table);
#ifndef SPOIL
	__WFI();	// go to sleep
#else
	for (;;)
	{
#ifdef REPAIR
		__disable_irq();
#endif
		LED_PORT->ODR |= 0;
#ifdef REPAIR
		__enable_irq();
#endif
	}
#endif
}
//========================================================================
// soft PWM data
#define PWM_CHANNELS	2

enum {CH_GREEN, CH_BLUE};	// channel names/functions
static const uint16_t pwm_channel_mask[PWM_CHANNELS] = {
	// channel port bit masks
	1 << GREEN_LED_BIT,
	1 << BLUE_LED_BIT
};
static uint8_t pwm_next_val[PWM_CHANNELS];	// pwm reload values
//========================================================================
// soft PWM routine, called from timer interrupt
static __INLINE _Bool softpwm(void)
{
	static uint8_t pwm_val[PWM_CHANNELS];	// current PWM values
	static uint32_t mask;
	static uint8_t pwmcnt;
	uint32_t i;
	
	LED_PORT->BSRR = mask;	// use previously set mask
	mask = 0;
	
	if (++ pwmcnt == PWM_STEPS)
	{
		// start of new period
		pwmcnt = 0;
		// set values & turn on active channels
		for (i = 0; i < PWM_CHANNELS; i ++)
			if ((pwm_val[i] = pwm_next_val[i]) != 0)
				mask |= pwm_channel_mask[i];
		return 1;
	}
	else
	{
		// turn off channels on pwm value match
		for (i = 0; i < PWM_CHANNELS; i ++)
			if (pwm_val[i] == pwmcnt)
				mask |= pwm_channel_mask[i];
		mask <<= 16;	// shift to get the reset mask
		return 0;
	}
}
//========================================================================
void TIM6_DAC_IRQHandler(void)
{
	static uint8_t blink_timer = BLINK_PERIOD;
	static enum {BM_SLOW, BM_FAST, BM_OFF} blink_mode = BM_SLOW;
	static uint8_t bstate = 0;
	static uint8_t blue_led_timer = 0, on_time = 0;
	static const uint8_t blink_periods[] = {100, 50, 1};
	static uint8_t tdiv = 0;
	static uint8_t pwm_target[PWM_CHANNELS] = {LED_DIM, LED_DIM};

	TIM6->SR = ~TIM_SR_UIF;	// clear interrupt flag

	if (softpwm())	// must be invoked at the very beginning of timer ISR !
	{
		// softpwm() returns 1 at start of PWM period
		uint32_t i;
		
		for (i = 0; i < PWM_CHANNELS; i ++)
			if (pwm_next_val[i] != pwm_target[i])
			{
				if (pwm_next_val[i] < pwm_target[i])
					pwm_next_val[i] ++;
				else
					pwm_next_val[i] --;
			}
				
	}

	if (++ tdiv == SOFTPWM_INT_FREQ / 100)
	{
		tdiv = 0;
		
		bstate = (bstate << 1 & 0xf) | (BUTTON_PORT->IDR >> BUTTON_BIT & 1);
		
		if (blue_led_timer)
		{
			if (-- blue_led_timer == 0)
				pwm_target[CH_BLUE] = LED_DIM;
		}
		else if (bstate == 1)
		{
			// button was released, now it is pressed - change blink mode
			if (++ blink_mode > BM_OFF)
				blink_mode = BM_SLOW;
			blue_led_timer = BTN_ACK_PERIOD;
			pwm_target[CH_BLUE] = LED_MAX;
		}
		
		if (-- blink_timer == 0)
		{
			blink_timer = blink_periods[blink_mode];
			on_time = blink_timer >> 1;
			pwm_target[CH_GREEN] = LED_DIM;
		}
		else if (blink_timer == on_time)
			pwm_target[CH_GREEN] = LED_MAX;
	}
}
